import numpy as np
import matplotlib.pyplot as plt
from copy import copy

""" Baut einen schoenen Latex String fuer Monome
"""
def fancy(xx, nn):
    ret = r''
    if abs(xx) > 0:
        if abs(xx) != 1 or nn == 0:
            ret += str(xx)
        if nn != 0:
            ret += r'x'
            if nn != 1:
                ret +=r'^{'+str(nn)+r'}'
    return ret

""" Eine Klasse fuer ein Polynom
"""
class mypolynomial:
    """ Hiermit wird der Datentyp erstellt, self ist der Verweis auf
        das Objekt, cc Koeffizienten
        Args:
        szu erstellendes mypolynomiall
        cc -- Koeffizienten des Polynoms
    """
    def __init__(self, cc):
        self.degree = len(cc)-1
        # effektiven Grad herausfinden
        while (self.degree >= 0 and cc[self.degree] == 0):
            self.degree -= 1
        # Koeffizienten speichern
        self.coefficients = np.array(cc[:self.degree+1])
        # Aus den Koeffizienten eine Stringdarstellung generieren
        if self.degree < 0:
            self.label = r'$0$'
        else:
            self.label = r'$'+fancy(self.coefficients[self.degree], self.degree)
            for ii in xrange(self.degree-1, -1, -1):
                if self.coefficients[ii] > 0:
                    self.label += '+'
                self.label += fancy(self.coefficients[ii], ii)
            self.label += r'$'

    """ Auswerten dieses Polynoms im Horner Schema
        Args:
        self -- Verweis auf Object vor '( )'
        xx -- Auswertestelle(n), auch array ok
    """
    def __call__(self, xx):
        # genug Nullen erstellen um NumPy Verwertung mit xx zu
        # erlauben
        ret = np.zeros_like(xx)
        # Horner Schema
        for ii in xrange(self.degree,-1,-1):
            ret = ret*xx+self.coefficients[ii]
        return ret

    """ Diese Funktion wird bei print und str verwendet, um eine
        \nTextdarstellung des Objekts zu bekommen
        Args:
        self -- mypolynomial, das in Text verwandelt werden soll
    """
    def __str__(self):
        return self.label

    """ Hiermit kann wie mit Listen mit [] auf die Koeffzienten zugegriffen werden
        Args:
        self -- Verweis auf mypolynomial vor '.'
        ii -- Index des gesuchten Koeffizienten
    """
    def __getitem__(self, ii):
        return self.coefficients[ii]

    """ Diese Funktion wird bei '+' aufgerufen
        Man beachte die Verwendung von copy wegen Python 'call by name'
        Args:
        self -- mypolynomial vor '+'
        other -- Objekt hinter '+'
    """
    def __add__(self, other):
        coeff = None
        # Falls auch ein Polynom, dann Koeffizienten addieren
        if isinstance(other, self.__class__):
            coeff = np.zeros(np.max([self.degree, other.degree])+1)
            coeff[:self.degree+1] = self.coefficients
            coeff[:other.degree+1] += other.coefficients
        # ansonsten als Skalar behandeln und auf den 0 Koeffzienten drauf
        else:
            coeff = self.coefficients.copy()
            coeff[0] += other
        return mypolynomial(coeff)

    """ Wir bei Additionen Aufgerufen, in denen das Polynom rechts
        steht und links keines
        Args:
        self -- mypolynomial hinter '+'
        other -- Nicht-mypolynomial vor '+'
    """
    def __radd(self, other):
        return self+other

    """ Unaeres -, verwendet bei (-p)
        Args:
        self -- mypolynomial hinter unaerem '-'
    """
    def __neg__(self):
        return mypolynomial(-self.coefficients)

    """ Bineares '-' mit Verwendung von unaerem
        Args:
        self -- mypolynomial vor '-'
        other -- Objekt hinter '-'
    """
    def __sub__(self, other):
        return self +(-other)
    
    """ Polynommultiplikation mit Polynom links von '*'
        Args:
        self -- mypolnomial vor '*'
        other -- Objekt hinter *
    """
    def __mul__(self, other):
        # Fuer Polynome die Koeffizienten zusammen multiplizieren
        if isinstance(other, self.__class__):
            # genuegend Platz erzeugen
            coeff = np.zeros(self.degree+other.degree+1)
            # Schleife ueber Grad
            for ii in xrange(len(coeff)):
                # Schleife fuer die Multiplikation, if Abfragen in
                # range gezogen
                for jj in xrange(max(0,ii-other.degree),min(ii, self.degree)+1):
                    coeff[ii] += self.coefficients[jj]*other.coefficients[ii-jj]
            return mypolynomial(coeff)
        # Ansonsten als Multiplikation mit Skalar deuten
        else:
            return mypolynomial(self.coefficients*other)

    """ Multiplikation mit Polynom rechts hinter '*'
        Args:
        self -- mypolynomial hinter '*'
        other -- Nicht-mypolynomial vor '*'
    """
    def __rmul__(self, other):
        return self*other

    """ Integral von aa bis bb auswerten als Polynomauswertung der Stammfunktion
        Args:
        self -- zu integrierendes mypolnomial
        aa -- linke Intervallgrenze
        bb -- rechte Intervallgrenze
    """
    def integrate(self, aa, bb):
        ret = np.zeros(2)
        bounds = np.array([aa, bb])
        # Horner Schema mit modifizieten Koeffizienten
        for ii in xrange(self.degree,-1,-1):
            ret = ret*bounds+self.coefficients[ii]/(ii+1.)
        # noch eine Multiplikation wegen 1 hoeherem Grad
        ret *= bounds
        return ret[1]-ret[0]

    """ Das ganze Ding auf [aa, bb] plotten.
        Malt ins aktuelle Bild rein.
        Default Argumente fuer nn und label gesetzt, man kann das also auch nur als
            pp.plot(aa, bb)
        benutzen.
        Args:
        self -- zu plottendes mypolynomial
        aa -- linke Intervallgrenze
        bb -- rechte Intervallgrenze
        nn -- Anzahl Auswertestellen
        label -- Beschriftung im Bild
    """
    def plot(self, aa, bb, nn=50, label=None):
        xx = np.linspace(aa, bb, nn)
        if label == None:
            label = self.label
        plt.plot(xx, self(xx), label=label)

""" Gram Schmidt mit der eigenen Polynomklasse zum Gewicht 1 auf [aa, bb]
    Args:
    polys -- zur orthonormalisierende Polynome
    aa -- linke Intervallgrenze
    bb -- rechte Intervallgrenze
"""
def ortho(polys, aa, bb):
    nn = len(polys)
    # leere Liste
    ret = copy(polys)
    # Schleife ueber die vorgegebenen Polynome
    for ii in xrange(nn):
        for jj in xrange(ii):
            ret[ii] = ret[ii] - (ret[ii]*ret[jj]).integrate(aa, bb)*ret[jj]
        ret[ii] = ret[ii] * (1./np.sqrt((ret[ii]*ret[ii]).integrate(aa, bb)))
    return ret

""" Baut aus vorgegebenen Polynomen auf [aa,bb] die Gram Matrix
    Args:
    polys -- Polynome
    aa -- linke Intervallgrenze
    bb -- rechte Intervallgrenze
"""
def get_gram(polys, aa, bb):
    nn = len(polys)
    # leere Matrix
    ret = np.zeros((nn, nn))
    # Matrix fuellen
    for ii in xrange(nn):
        ret[ii, ii] = (polys[ii]*polys[ii]).integrate(aa, bb)
        for jj in xrange(ii+1, nn):
            ret[ii, jj] = (polys[ii]*polys[jj]).integrate(aa, bb)
            ret[jj, ii] = ret[ii, jj]
    return ret

if __name__ == '__main__':
    # ein paar Polynome erstellen
    p0 = mypolynomial([1])
    p1 = mypolynomial([0, 1])
    p2 = p1*p1
    p3 = p2*p1
    pp = 2*p3-3*p2+1 
    pt = (p1 - 1)*(p1 + 1)
    # Auswertetests
    print 'Evaluation tests', pp(0), pp(1), pp(np.array([0, 1]))

    # Ein paar Plots
    fig = plt.figure()
    p0.plot(0, 1)
    p1.plot(0, 1)
    p2.plot(0, 1)
    pp.plot(0, 1)
    pt.plot(0, 1)
    plt.gca().legend()
    fig.show()

    # Integrale ausrechnen testen
    print 'Integral tests', p0.integrate(0, 1), p1.integrate(0, 1), p2.integrate(0, 1), pp.integrate(0, 1)

    # Orthogonalisierung testen
    degree = 4
    monomials = [mypolynomial([0]*ii+[1]) for ii in xrange(degree+1)]
    leg = ortho(monomials, -1, 1)
    # Orthonormierte Polynome testen
    fig=plt.figure()
    print 'Orthonormalized'
    for ii in xrange(degree+1):
        leg[ii].plot(-1, 1)
        print '    ', leg[ii]
    plt.gca().legend()
    fig.show()

    # Gram Matrizen
    AA = get_gram(monomials, 0, 1)
    print AA
    BB = get_gram(leg, -1, 1)
    print BB

